@@ -7,19 +7,26 @@ from http import HTTPStatus
 from typing import Any, Dict, List, Optional, Union
 
 from aiohttp import ClientConnectionError, ClientSession
+from multidict import CIMultiDict, CIMultiDictProxy
 
 from .const import (
     BASE_URL_V4,
     CURRENT,
     DAILY,
+    FIFTEEN_MINUTES,
+    FIVE_MINUTES,
     FORECASTS,
     HEADER_DAILY_API_LIMIT,
+    HEADER_REMAINING_CALLS_IN_SECOND,
     HEADERS,
     HOURLY,
+    MAX_FIELDS_PER_REQUEST,
     NOWCAST,
     ONE_DAY,
     ONE_HOUR,
-    REALTIME,
+    ONE_MINUTE,
+    THIRTY_MINUTES,
+    TIMESTEP_CURRENT,
     TIMESTEP_DAILY,
     TIMESTEP_HOURLY,
     VALID_TIMESTEPS,
@@ -70,6 +77,26 @@ def dt_to_utc(input_dt: datetime) -> datetime:
     return input_dt
 
 
+def _timedelta_to_str(timestep: timedelta) -> str:
+    """Convert timedelta to timestep string."""
+    if timestep == ONE_DAY:
+        return TIMESTEP_DAILY
+    if timestep == ONE_HOUR:
+        return TIMESTEP_HOURLY
+    if timestep not in (THIRTY_MINUTES, FIFTEEN_MINUTES, FIVE_MINUTES, ONE_MINUTE):
+        raise InvalidTimestep(f"Invalid `timestep` value {timestep}")
+    return f"{int(timestep.total_seconds()/60)}m"
+
+
+def _timestep_to_key(timestep: str) -> str:
+    """Convert timestep to dict key."""
+    if timestep == TIMESTEP_DAILY:
+        return DAILY
+    if timestep == TIMESTEP_HOURLY:
+        return HOURLY
+    return NOWCAST
+
+
 class TomorrowioV4:
     """Async class to query the Tomorrow.io v4 API."""
 
@@ -82,12 +109,8 @@ class TomorrowioV4:
         session: Optional[ClientSession] = None,
     ) -> None:
         """Initialize Tomorrow.io API object."""
-        if unit_system.lower() not in ("metric", "imperial", "si", "us"):
+        if unit_system.lower() not in ("metric", "imperial"):
             raise ValueError("`unit_system` must be `metric` or `imperial`")
-        if unit_system.lower() == "si":
-            unit_system = "metric"
-        elif unit_system.lower() == "us":
-            unit_system = "imperial"
 
         self._apikey = apikey
         self.unit_system = unit_system.lower()
@@ -97,12 +120,23 @@ class TomorrowioV4:
             "units": self.unit_system,
         }
         self._headers = {**HEADERS, "apikey": self._apikey}
-        self._max_requests_per_day: Optional[int] = None
+        self._rate_limits: CIMultiDict = CIMultiDict()
+        self._num_api_requests: int = 0
+
+    @property
+    def rate_limits(self) -> CIMultiDictProxy:
+        """Return tomorrow.io rate limits for API key"""
+        return CIMultiDictProxy(self._rate_limits)  # make read-only
 
     @property
     def max_requests_per_day(self) -> Optional[int]:
         """Return the maximum number of requests per day."""
-        return self._max_requests_per_day
+        return self.rate_limits.get(HEADER_DAILY_API_LIMIT)
+
+    @property
+    def num_api_requests(self) -> int:
+        """The number of API requests made during the most recent call."""
+        return self._num_api_requests
 
     @staticmethod
     def convert_fields_to_measurements(fields: List[str]) -> List[str]:
@@ -133,8 +167,12 @@ class TomorrowioV4:
 
         return fields
 
+    @staticmethod
+    def _get_url() -> str:
+        # This method is required for test mocks
+        return BASE_URL_V4
+
     async def _call_api(self, params: Dict[str, Any]) -> Dict[str, Any]:
-        """Call Tomorrow.io API."""
         if self._session:
             return await self._make_call(params, self._session)
 
@@ -144,9 +182,18 @@ class TomorrowioV4:
     async def _make_call(
         self, params: Dict[str, Any], session: ClientSession
     ) -> Dict[str, Any]:
+        try:
+            if (
+                self._rate_limits
+                and self._rate_limits[HEADER_REMAINING_CALLS_IN_SECOND] == 0
+            ):
+                await asyncio.sleep(1)
+        except LookupError as error:
+            raise UnknownException("Missing Header") from error
+
         try:
             resp = await session.post(
-                BASE_URL_V4,
+                self._get_url(),
                 headers=self._headers,
                 data=json.dumps({**self._params, **params}),
             )
@@ -154,10 +201,24 @@ class TomorrowioV4:
         except ClientConnectionError as error:
             raise CantConnectException() from error
 
+        self._rate_limits = CIMultiDict(
+            {k: int(v) for k, v in resp.headers.items() if "ratelimit" in k.lower()}
+        )
+
         if resp.status == HTTPStatus.OK:
-            max_requests = resp.headers[HEADER_DAILY_API_LIMIT]
-            if max_requests != self._max_requests_per_day:
-                self._max_requests_per_day = int(max_requests)
+            self._num_api_requests += 1
+            for warning in set(
+                warning["message"] for warning in resp_json.get("warnings", [])
+            ):
+                _LOGGER.warning(
+                    (
+                        "While calling the API for the following timesteps: [%s], the "
+                        "following warning was returned: %s"
+                    ),
+                    ", ".join(params["timesteps"]),
+                    warning,
+                )
+
             return resp_json
         if resp.status == HTTPStatus.BAD_REQUEST:
             raise MalformedRequestException(resp_json, resp.headers)
@@ -165,54 +226,81 @@ class TomorrowioV4:
             raise InvalidAPIKeyException(resp_json, resp.headers)
         if resp.status == HTTPStatus.TOO_MANY_REQUESTS:
             raise RateLimitedException(resp_json, resp.headers)
+
+        resp.raise_for_status()
         raise UnknownException(resp_json, resp.headers)
 
-    async def realtime(self, fields: List[str]) -> Dict[str, Any]:
+    async def realtime(
+        self, fields: List[str], reset_num_api_requests: bool = True
+    ) -> Dict[str, Any]:
         """Return realtime weather conditions from Tomorrow.io API."""
-        return await self._call_api(
-            {
-                "fields": process_v4_fields(fields, REALTIME),
-                "timesteps": ["current"],
-            }
-        )
+        if reset_num_api_requests:
+            self._num_api_requests = 0
 
-    async def _forecast(
+        ret_data = {}
+        for i in range(0, len(fields), MAX_FIELDS_PER_REQUEST):
+            data = await self._call_api(
+                {
+                    "timesteps": [TIMESTEP_CURRENT],
+                    "fields": fields[i : i + MAX_FIELDS_PER_REQUEST],
+                }
+            )
+            try:
+                ret_data.update(data["data"]["timelines"][0]["intervals"][0]["values"])
+            except LookupError as error:
+                raise UnknownException(data) from error
+
+        return ret_data
+
+    async def forecast(
         self,
-        timestep: timedelta,
+        timesteps: List[timedelta],
         fields: List[str],
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
+        reset_num_api_requests: bool = True,
         **kwargs,
-    ) -> Dict[str, Any]:
+    ) -> Dict[str, List[Dict[str, Any]]]:
         """Return forecast data from Tomorrow.io's API for a given time period."""
-        if timestep not in VALID_TIMESTEPS:
-            raise InvalidTimestep(f"{timestep} is not a valid 'timestep' parameter")
-        fields = process_v4_fields(fields, timestep)
-        if timestep > ONE_HOUR:
-            fields = self.convert_fields_to_measurements(fields)
+        if reset_num_api_requests:
+            self._num_api_requests = 0
 
         params: Dict[str, Any] = {
-            "fields": fields,
+            "timesteps": [_timedelta_to_str(timestep) for timestep in timesteps],
             **kwargs,
         }
-        if timestep == ONE_DAY:
-            params["timesteps"] = [TIMESTEP_DAILY]
-        elif timestep == ONE_HOUR:
-            params["timesteps"] = [TIMESTEP_HOURLY]
-        else:
-            params["timesteps"] = [f"{int(timestep.total_seconds()/60)}m"]
 
         if start_time:
             if not start_time.tzinfo:
                 start_time.replace(tzinfo=timezone.utc)
-            params["startTime"] = start_time.replace(microsecond=0).isoformat()
         else:
             start_time = datetime.now(tz=timezone.utc)
-        if duration:
-            end_time = (start_time + duration).replace(microsecond=0)
-            params["endTime"] = end_time.isoformat()
+        params["startTime"] = start_time.replace(microsecond=0).isoformat()
 
-        return await self._call_api(params)
+        if duration:
+            params["endTime"] = (
+                (start_time + duration).replace(microsecond=0).isoformat()
+            )
+
+        forecasts: Dict[str, List[Dict[str, Any]]] = {}
+        for i in range(0, len(fields), MAX_FIELDS_PER_REQUEST):
+            data = await self._call_api(
+                {**params, "fields": fields[i : i + MAX_FIELDS_PER_REQUEST]}
+            )
+            try:
+                for timeline in data["data"]["timelines"]:
+                    forecast_type = _timestep_to_key(timeline["timestep"])
+                    if forecast_type not in forecasts:
+                        forecasts[forecast_type] = timeline["intervals"]
+                        continue
+                    for idx in range(0, len(forecasts[forecast_type])):
+                        forecasts[forecast_type][idx]["values"].update(
+                            timeline["intervals"][idx]["values"]
+                        )
+            except LookupError as error:
+                raise UnknownException(data) from error
+
+        return forecasts
 
     async def forecast_nowcast(
         self,
@@ -220,120 +308,130 @@ class TomorrowioV4:
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
         timestep: int = 5,
-    ) -> Dict[str, Any]:
+        reset_num_api_requests: bool = True,
+    ) -> List[Dict[str, Any]]:
         """Return forecast data from Tomorrow.io's NowCast API for a given time period."""
-        if timestep not in (1, 5, 15, 30):
-            raise InvalidTimestep
-        return await self._forecast(
-            timedelta(minutes=timestep),
+        forecasts = await TomorrowioV4.forecast(
+            self,
+            [timedelta(minutes=timestep)],
             fields,
             start_time=start_time,
             duration=duration,
+            reset_num_api_requests=reset_num_api_requests,
         )
+        return forecasts[NOWCAST]
 
     async def forecast_daily(
         self,
         fields: List[str],
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
-    ) -> Dict[str, Any]:
+        reset_num_api_requests: bool = True,
+    ) -> List[Dict[str, Any]]:
         """Return daily forecast data from Tomorrow.io's API for a given time period."""
-        return await self._forecast(
-            ONE_DAY, fields, start_time=start_time, duration=duration
+        forecasts = await TomorrowioV4.forecast(
+            self,
+            [ONE_DAY],
+            fields,
+            start_time=start_time,
+            duration=duration,
+            reset_num_api_requests=reset_num_api_requests,
         )
+        return forecasts[DAILY]
 
     async def forecast_hourly(
         self,
         fields: List[str],
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
-    ) -> Dict[str, Any]:
+        reset_num_api_requests: bool = True,
+    ) -> List[Dict[str, Any]]:
         """Return hourly forecast data from Tomorrow.io's API for a given time period."""
-        return await self._forecast(
-            ONE_HOUR, fields, start_time=start_time, duration=duration
+        forecasts = await TomorrowioV4.forecast(
+            self,
+            [ONE_HOUR],
+            fields,
+            start_time=start_time,
+            duration=duration,
+            reset_num_api_requests=reset_num_api_requests,
+        )
+        return forecasts[HOURLY]
+
+    async def all_forecasts(
+        self,
+        fields: List[str],
+        start_time: Optional[datetime] = None,
+        duration: Optional[timedelta] = None,
+        nowcast_timestep: int = 5,
+        reset_num_api_requests: bool = True,
+    ) -> Dict[str, List[Dict[str, Any]]]:
+        """Return all forecasts."""
+        return await TomorrowioV4.forecast(
+            self,
+            [
+                timedelta(minutes=nowcast_timestep),
+                ONE_HOUR,
+                ONE_DAY,
+            ],
+            fields,
+            start_time=start_time,
+            duration=duration,
+            reset_num_api_requests=reset_num_api_requests,
         )
 
     async def realtime_and_all_forecasts(
         self,
         realtime_fields: List[str],
-        forecast_or_nowcast_fields: List[str],
-        hourly_fields: List[str] = None,
-        daily_fields: List[str] = None,
+        all_forecasts_fields: Optional[List[str]] = None,
+        nowcast_fields: Optional[List[str]] = None,
+        hourly_fields: Optional[List[str]] = None,
+        daily_fields: Optional[List[str]] = None,
         nowcast_timestep: int = 5,
-    ) -> Dict[str, Any]:
+    ) -> Dict[str, Dict[str, Any]]:
         """
         Return realtime weather and all forecasts.
 
-        If `hourly_fields` and `daily_fields` are not provided,
-        `forecast_or_nowcast_fields` will be used to get nowcast, hourly, and daily
-        data.
+        To get the same fields for all forecasts, use all_forecasts_fields. To get
+        specific fields for specific forecast types, use the corresponding fields list.
         """
-        ret_data = {}
-        data = await self._call_api(
-            {
-                "timesteps": ["current"],
-                "fields": realtime_fields,
-            }
-        )
-        if (
-            "data" in data
-            and "timelines" in data["data"]
-            and "intervals" in data["data"]["timelines"][0]
-            and "values" in data["data"]["timelines"][0]["intervals"][0]
+        self._num_api_requests = 0
+        if not (
+            all_forecasts_fields or nowcast_fields or hourly_fields or daily_fields
         ):
-            ret_data[CURRENT] = data["data"]["timelines"][0]["intervals"][0]["values"]
+            raise ValueError("At least one field list must be specified")
+        if all_forecasts_fields and any(
+            fields for fields in (nowcast_fields, hourly_fields, daily_fields)
+        ):
+            raise ValueError(
+                "Either only all_forecasts_fields list must be specified or at least "
+                "one of the other field lists"
+            )
 
-        forecasts = ret_data.setdefault(FORECASTS, {})
-        start_time = datetime.now(tz=timezone.utc).isoformat()
-
-        if not hourly_fields and not daily_fields:
-            data = await self._call_api(
-                {
-                    "timesteps": [
-                        f"{nowcast_timestep}m",
-                        TIMESTEP_HOURLY,
-                        TIMESTEP_DAILY,
-                    ],
-                    "fields": forecast_or_nowcast_fields,
-                    "startTime": start_time,
-                }
+        forecasts: Dict[str, List[Dict[str, Any]]] = {}
+        if all_forecasts_fields:
+            forecasts = await TomorrowioV4.all_forecasts(
+                self,
+                all_forecasts_fields,
+                nowcast_timestep=nowcast_timestep,
+                reset_num_api_requests=False,
             )
-            if "data" in data and "timelines" in data["data"]:
-                for timeline in data["data"]["timelines"]:
-                    if timeline["timestep"] == TIMESTEP_DAILY:
-                        key = DAILY
-                    elif timeline["timestep"] == TIMESTEP_HOURLY:
-                        key = HOURLY
-                    else:
-                        key = NOWCAST
-                    forecasts[key] = timeline["intervals"]
         else:
-            data = await self._call_api(
-                {
-                    "timesteps": [f"{nowcast_timestep}m"],
-                    "fields": forecast_or_nowcast_fields,
-                    "startTime": start_time,
-                }
-            )
-            if "data" in data and "timelines" in data["data"]:
-                forecasts[NOWCAST] = data["data"]["timelines"][0]["intervals"]
-
-            for field_list, timestep, key in (
-                (hourly_fields, TIMESTEP_HOURLY, HOURLY),
-                (daily_fields, TIMESTEP_DAILY, DAILY),
-            ):
-                if field_list:
-                    await asyncio.sleep(1)
-                    data = await self._call_api(
-                        {
-                            "timesteps": [timestep],
-                            "fields": field_list,
-                            "startTime": start_time,
-                        }
+            for fields, timestep in [
+                (nowcast_fields, timedelta(minutes=nowcast_timestep)),
+                (hourly_fields, ONE_HOUR),
+                (daily_fields, ONE_DAY),
+            ]:
+                if fields:
+                    forecasts.update(
+                        await TomorrowioV4.forecast(
+                            self, [timestep], fields, reset_num_api_requests=False
+                        )
                     )
-                    if "data" in data and "timelines" in data["data"]:
-                        forecasts[key] = data["data"]["timelines"][0]["intervals"]
-        return ret_data
+
+        current = await TomorrowioV4.realtime(
+            self, realtime_fields, reset_num_api_requests=False
+        )
+        return {CURRENT: current, FORECASTS: forecasts}
 
 
 class TomorrowioV4Sync(TomorrowioV4):
@@ -350,9 +448,33 @@ class TomorrowioV4Sync(TomorrowioV4):
         super().__init__(apikey, latitude, longitude, unit_system)
 
     @async_to_sync
-    async def realtime(self, fields: List[str]) -> Dict[str, Any]:
+    async def realtime(
+        self, fields: List[str], reset_num_api_requests: bool = True
+    ) -> Dict[str, Any]:
         """Return realtime weather conditions from Tomorrow.io API."""
-        return await super().realtime(fields)
+        return await super().realtime(
+            fields, reset_num_api_requests=reset_num_api_requests
+        )
+
+    @async_to_sync
+    async def forecast(
+        self,
+        timesteps: List[timedelta],
+        fields: List[str],
+        start_time: Optional[datetime] = None,
+        duration: Optional[timedelta] = None,
+        reset_num_api_requests: bool = True,
+        **kwargs,
+    ) -> Dict[str, List[Dict[str, Any]]]:
+        """Return forecast data from Tomorrow.io's API for a given time period."""
+        return await super().forecast(
+            timesteps,
+            fields,
+            start_time=start_time,
+            duration=duration,
+            reset_num_api_requests=reset_num_api_requests,
+            **kwargs,
+        )
 
     @async_to_sync
     async def forecast_nowcast(
@@ -361,9 +483,16 @@ class TomorrowioV4Sync(TomorrowioV4):
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
         timestep: int = 5,
-    ) -> Dict[str, Any]:
+        reset_num_api_requests: bool = True,
+    ) -> List[Dict[str, Any]]:
         """Return forecast data from Tomorrow.io's NowCast API for a given time period."""
-        return await super().forecast_nowcast(fields, start_time, duration, timestep)
+        return await super().forecast_nowcast(
+            fields,
+            start_time=start_time,
+            duration=duration,
+            timestep=timestep,
+            reset_num_api_requests=reset_num_api_requests,
+        )
 
     @async_to_sync
     async def forecast_daily(
@@ -371,9 +500,15 @@ class TomorrowioV4Sync(TomorrowioV4):
         fields: List[str],
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
-    ) -> Dict[str, Any]:
+        reset_num_api_requests: bool = True,
+    ) -> List[Dict[str, Any]]:
         """Return daily forecast data from Tomorrow.io's API for a given time period."""
-        return await super().forecast_daily(fields, start_time, duration)
+        return await super().forecast_daily(
+            fields,
+            start_time=start_time,
+            duration=duration,
+            reset_num_api_requests=reset_num_api_requests,
+        )
 
     @async_to_sync
     async def forecast_hourly(
@@ -381,29 +516,54 @@ class TomorrowioV4Sync(TomorrowioV4):
         fields: List[str],
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
-    ) -> Dict[str, Any]:
+        reset_num_api_requests: bool = True,
+    ) -> List[Dict[str, Any]]:
         """Return hourly forecast data from Tomorrow.io's API for a given time period."""
-        return await super().forecast_hourly(fields, start_time, duration)
+        return await super().forecast_hourly(
+            fields,
+            start_time=start_time,
+            duration=duration,
+            reset_num_api_requests=reset_num_api_requests,
+        )
+
+    @async_to_sync
+    async def all_forecasts(
+        self,
+        fields: List[str],
+        start_time: Optional[datetime] = None,
+        duration: Optional[timedelta] = None,
+        nowcast_timestep: int = 5,
+        reset_num_api_requests: bool = True,
+    ) -> Dict[str, List[Dict[str, Any]]]:
+        """Return all forecasts."""
+        return await super().all_forecasts(
+            fields,
+            start_time=start_time,
+            duration=duration,
+            nowcast_timestep=nowcast_timestep,
+            reset_num_api_requests=reset_num_api_requests,
+        )
 
     @async_to_sync
     async def realtime_and_all_forecasts(
         self,
         realtime_fields: List[str],
-        forecast_or_nowcast_fields: List[str],
-        hourly_fields: List[str] = None,
-        daily_fields: List[str] = None,
+        all_forecasts_fields: Optional[List[str]] = None,
+        nowcast_fields: Optional[List[str]] = None,
+        hourly_fields: Optional[List[str]] = None,
+        daily_fields: Optional[List[str]] = None,
         nowcast_timestep: int = 5,
-    ) -> Dict[str, Any]:
+    ) -> Dict[str, Dict[str, Any]]:
         """
         Return realtime weather and all forecasts.
 
-        If `hourly_fields` and `daily_fields` are not provided,
-        `forecast_or_nowcast_fields` will be used to get nowcast, hourly, and daily
-        data.
+        To get the same fields for all forecasts, use all_forecasts_fields. To get
+        specific fields for specific forecast types, use the corresponding fields list.
         """
         return await super().realtime_and_all_forecasts(
             realtime_fields=realtime_fields,
-            forecast_or_nowcast_fields=forecast_or_nowcast_fields,
+            all_forecasts_fields=all_forecasts_fields,
+            nowcast_fields=nowcast_fields,
             hourly_fields=hourly_fields,
             daily_fields=daily_fields,
             nowcast_timestep=nowcast_timestep,
